AttachSpec("package/spec");

/* A covariant (U, V)^level */
COV_t :=  recformat<
    type,               // 0 (multiplitive) or 1 (additive)
    U, 			// Covariants in U
    V, 			// Covariants in V
    level,		// Transvectant order
    degree,             // Covariant degree
    order               // Covariant order
    >;

FF := RationalField();
_<_J10, _J9, _J8, _J7, _J6, _J5, _J4, _J3, _J2 > := PolynomialRing(FF, [10, 9, 8, 7, 6, 5, 4, 3, 2]);

load "isocovgenus3.dat"; FdCov := Covariants;

/* Transvectants  */
function UnivariateTransvectant(F, G, r)

    Q, Qdeg, n := Explode(F);
    R, Rdeg, m := Explode(G);
    
    n := IntegerRing()!n;
    m := IntegerRing()!m;
    
    K := BaseRing(Parent(Q));
    
    h := Parent(Q)!0;
    for k := 0 to r do
	h +:= (-1)^k
	    * Binomial(m-k,r-k)  * Derivative(Q, r-k)
	    * Binomial(n-r+k, k) * Derivative(R, k);
    end for;

    coef := Factorial(r)*Factorial(m-r)*Factorial(n-r)/Factorial(m)/Factorial(n);
    h := (K!coef) * h;
    
    return [* h, Qdeg+Rdeg, n+m-2*r *];
    
end function;

/* Computing the Covariant "Cov" for the form "form" thanks to the precomputed table "FdCov"  */
forward GetCovariant;
function GetCovariant(Cov, FdCov, form)

    if Cov`type eq 0 then
	
	/* Is Cov equal to the initial form ? */
	if (Cov`U eq {* *}  and Cov`V eq {* 0 *}) or
	    (Cov`U eq {* 0 *}  and Cov`V eq {* *}) then
	    return [form, 1];
	end if;

	/* First, let us obtain the covariant U_cov */
	U_cov := 0*form + 1; U_deg := 0; U_ord := 0;
	for cov_idx in MultisetToSet(Cov`U) do
	    
	    cov, _ := Explode(GetCovariant(FdCov[cov_idx], FdCov, form));
	    
	    U_cov *:= cov ^ Multiplicity(Cov`U, cov_idx);
	    U_deg +:= FdCov[cov_idx]`degree * Multiplicity(Cov`U, cov_idx);
	    U_ord +:= FdCov[cov_idx]`order * Multiplicity(Cov`U, cov_idx);
	    
	end for;
    
	/* Then, let us obtain the covariant V_cov */
	V_cov := 0*form + 1; V_deg := 0; V_ord := 0;
	for cov_idx in MultisetToSet(Cov`V) do
	    
	    cov, _ := Explode(GetCovariant(FdCov[cov_idx], FdCov, form));
	    
	    V_cov *:= cov ^ Multiplicity(Cov`V, cov_idx);
	    V_deg +:= FdCov[cov_idx]`degree * Multiplicity(Cov`V, cov_idx);
	    V_ord +:= FdCov[cov_idx]`order * Multiplicity(Cov`V, cov_idx);
	    
	end for;
    
	/* Output the transvectant */
	return UnivariateTransvectant([U_cov, U_deg, U_ord], [V_cov, V_deg, V_ord], Cov`level);
    end if;
    
    /* First, let us obtain the covariant U_cov */
    U_cov := 0*form; U_deg := 0; U_ord := 0;
    for cov_idx in MultisetToSet(Cov`U) do
	    
	cov, _ := Explode(GetCovariant(FdCov[cov_idx], FdCov, form));
	    
	U_cov +:= cov * Multiplicity(Cov`U, cov_idx);
	U_deg := Max(U_deg, FdCov[cov_idx]`degree);
	U_ord := Max(U_ord, FdCov[cov_idx]`order);
	    
    end for;
    
    return [*U_cov, U_deg, U_ord*];
    
end function;


procedure PrettyPrintCov(Cov, Covariants : expand := true)

    if expand eq false then
	printf "C_{%o,%o}", Cov`degree, Cov`order;
	return;
    end if;

    /* Is Cov equal to the initial form ? */
    if (Cov`U eq {* *}  and Cov`V eq {* 0 *}) or
	(Cov`U eq {* 0 *}  and Cov`V eq {* *}) then
	printf "f";
	return;
    end if;

    printf "(";
      
    /* First, let us obtain the covariant U_cov */
    for cov_idx in MultisetToSet(Cov`U) do
	PrettyPrintCov(Covariants[cov_idx], Covariants : expand := expand);
	if Multiplicity(Cov`U, cov_idx) ne 1 then
	    printf "^%o ", Multiplicity(Cov`U, cov_idx);
	else
	    printf " ";
	end if;
    end for;
    
    printf ",";
	  
    /* Then, let us obtain the covariant V_cov */
    for cov_idx in MultisetToSet(Cov`V) do
	PrettyPrintCov(Covariants[cov_idx], Covariants : expand := expand);
	if Multiplicity(Cov`V, cov_idx) ne 1 then
	    printf "^%o ", Multiplicity(Cov`V, cov_idx);
	else
	    printf " ";
	end if;
    end for;

    if Cov`level gt 9 then
	printf ")_{%o}", Cov`level;
    else
	printf ")_%o", Cov`level;
    end if;

end procedure;

function QuarticInvariants(k_cov)
    e, d, c, b, a := Explode([Coefficient(k_cov[1], i) : i in [0..4]]);
    I := 12*a*e - 3*b*d + c^2;
    J := 72*a*c*e + 9*b*c*d +-27*a*d^2 - 27*e*b^2 - 2*c^3;
    return [* [* I, 2*k_cov[2], 0 *],  [* J, 3*k_cov[2], 0 *] *];
end function;

function TransformPolynomial(f,n,mat)
    // Transforms f, considered as a homogeneous polynomial of 
    // degree n, by the matrix with entries mat = [a,b,c,d].
    a,b,c,d := Explode(mat);
    x1 := a*Parent(f).1 + b; z1 := c*Parent(f).1 + d;
    return &+[Coefficient(f,i)*x1^i*z1^(n-i) : i in [0..n]];
end function;

function IsomorphismsWithCovariants(f, df, C, dC)
    SF := SplittingField(C); PP<X> := PolynomialRing(SF);

    ret, ML := IsGL2Equivalent(PP!C, PP!C, dC);

    /* Let us check that all the automorphisms work */
    RL := []; //ML;
    for M in ML do
	f2 := TransformPolynomial(PP!f, df, M);
	f2 := f2/LeadingCoefficient(f2);
	if f2 ne f then
//	    "HUM...", M, "DOES not work..";
	    continue;
	end if;
	Append(~RL, M);
    end for;

    return RL;

end function;


procedure CoPrint(IO, FdCov, f : expand := false)
    for ID in IO do
	PrettyPrintCov(FdCov[ID[1]], Covariants : expand := false); printf "=";
	if expand then
	    PrettyPrintCov(FdCov[ID[1]], Covariants : expand := true); printf "=";
	end if;
	CV :=  GetCovariant(FdCov[ID[1]], FdCov, f)[1];
	if CV eq 0 then printf "0"; else
	    printf "(%o) ",  Coefficient(CV, Degree(CV));
	    for cf in Factorization(CV) do
		printf "(%o)^%o ", cf[1], cf[2];
	    end for;
	end if;
	printf "\n";
    end for;
end procedure;

procedure CoPrintAll(FdCov, f : expand := false)
    
    // Fund. covariants of order 2
    ""; "Covariants of order 2 are :";
    IO := [[idx, FdCov[idx]`order] : idx in [1..#FdCov] |
	FdCov[idx]`order eq 2];
    CoPrint(IO, FdCov, f : expand := expand);

    // Fund. covariants of order 4
    ""; "Covariants of order 4 are :";
    IO := [[idx, FdCov[idx]`order] : idx in [1..#FdCov] |
	FdCov[idx]`order eq 4];
    CoPrint(IO, FdCov, f : expand := expand);

    // Fund. covariants of order 6
    ""; "Covariants of order 6 are :";
    IO := [[idx, FdCov[idx]`order] : idx in [1..#FdCov] |
	FdCov[idx]`order eq 6];
    CoPrint(IO, FdCov, f : expand := expand);

    // Non null covariants
    ""; "None zero covariants are :";
    IO := Sort([[idx, FdCov[idx]`order] : idx in [1..#FdCov] | GetCovariant(FdCov[idx], FdCov, f)[1] ne 0], func<x, y | x[2]-y[2]>);
    CoPrint(IO, FdCov, f : expand := expand);

    // Non degenerate covariants
    ""; "None degenerate covariants are :";
    IO := [[idx, FdCov[idx]`order] : idx in [1..#FdCov] |
	Degree(GetCovariant(FdCov[idx], FdCov, f)[1]) ge Max(1,FdCov[idx]`order-1) and
	Discriminant(GetCovariant(FdCov[idx], FdCov, f)[1]) ne 0 ];
    IO := Sort(IO, func<x, y | x[2]-y[2]>);
    CoPrint(IO, FdCov, f : expand := expand);
    
    printf "\n";

end procedure;


/*********************************************************************************************/
/************************* Dimension 0 *******************************************************/
/*********************************************************************************************/


/* Case C2 x S4 (48 elts)
 *************************/
""; "**************** C2 x S4 test";
FF := Rationals(); _<x> := PolynomialRing(FF);
f := x^8 + 14*x^4 + 1;
"Testing f =", f;
"Group", IdentifyGroup(GeometricAutomorphismGroupFromGenus3HyperellipticPolynomials(f));

CoPrintAll(FdCov, f : expand := true);

    
/* Case V8 (32 elts)
 ********************/
""; "**************** V8 test";
FF := Rationals(); _<x> := PolynomialRing(FF);
f := x^8 - 1;
"Testing f =", f;
"Group", IdentifyGroup(GeometricAutomorphismGroupFromGenus3HyperellipticPolynomials(f));

CoPrintAll(FdCov, f);

/* Case U6 (24 elts)
 *******************/
""; "**************** U6 test";
FF := Rationals(); _<x> := PolynomialRing(FF);
f := x * (x^6 - 1);
"Testing f =", f;
"Group", IdentifyGroup(GeometricAutomorphismGroupFromGenus3HyperellipticPolynomials(f));

CoPrintAll(FdCov, f);

/* Let us test the cov. of deg. 6... (a curve of genus 2 with aut. group of size 24)*/
C := GetCovariant(FdCov[11], FdCov, f)[1];
ISOs := IsomorphismsWithCovariants(f, 8, C, 6);
#ISOs, "isomorphisms found";
if #ISOs ne 12 then "ARGHHHH"; quit; end if;


/* Case C14 (14 elts)
 *********************/
""; "**************** C14 test";
FF := Rationals(); _<x> := PolynomialRing(FF);
f := x^7 - 1;
"Testing f =", f;
"Group", IdentifyGroup(GeometricAutomorphismGroupFromGenus3HyperellipticPolynomials(f));

CoPrintAll(FdCov, f);

/*********************************************************************************************/
/************************* Dimension 1 *******************************************************/
/*********************************************************************************************/


/* Case C2xD8 (16 elts)
 *********************/
""; "**************** C2xD8 test";

// Enumerating them
_<a> := FunctionField(Rationals(), 1); _<x> := PolynomialRing(Parent(a));

// Singular for a = +/-2 and C2xS4 group for ?
f := x^8 + a*x^4  + 1;
"Testing f =", f;
"Group", IdentifyGroup(GeometricAutomorphismGroupFromGenus3HyperellipticPolynomials(f));

// but none non-dgenerate covariant is of order smaller than 8...
CoPrintAll(FdCov, f);

/* D12 (12 elts)
 ****************/
""; "**************** D12 test";

// Enumerating them
_<a> := FunctionField(Rationals(), 1); _<x> := PolynomialRing(Parent(a));

// Singular for a = +/- 2 and C2xS4 group for a^2 + 49/8 = 0
f := x*(x^6 + a*x^3 + 1);
"Testing f =", f;
"Group", IdentifyGroup(GeometricAutomorphismGroupFromGenus3HyperellipticPolynomials(f));


// Generically, the 11 cov. of order 6 are non zeroes.
CoPrintAll(FdCov, f);


// The cov. [ 11, 6 ], x^6 + 48*t + 8/81
C := GetCovariant(FdCov[11], FdCov, f);
C, Factorization(Numerator(C[1]));

// It has a singularity only when a^2 + 49/8 = 0 (C2xS4 curves)
Factorization(Numerator(Discriminant(C[1])));

_<x> := PolynomialRing(Rationals());
_<a> := quo<PolynomialRing(Rationals()) | x^2 + 49/8>;

_<x> := PolynomialRing(Parent(a));
f := x*(x^6 + a*x^3 + 1);
IdentifyGroup(GeometricAutomorphismGroupFromGenus3HyperellipticPolynomials(f));

// Few tests
_<X> := PolynomialRing(Rationals());
for i := 1 to 100 do
    repeat
	A := Random(-1000, 1000);
	g := X*(X^6 + A*X^3 + 1);
    until IdentifyGroup(GeometricAutomorphismGroupFromGenus3HyperellipticPolynomials(g)) eq <12,4>;
    C := GetCovariant(FdCov[11], FdCov, g);
    ISOs := IsomorphismsWithCovariants(g, 8, C[1], C[3]);
//    #ISOs, "isomorphisms found";
    if #ISOs ne 6 then "ARGHHHH at", f; quit; end if;
end for;


/* C2xC4 (8 elts)
 ****************/
""; "**************** C2xC4 test";
    
// Enumerating them
_<a> := FunctionField(Rationals(), 1); _<x> := PolynomialRing(Parent(a));

// Singular for a = +/-2
f := (x^4-1)*(x^4 + a*x^2 +1);
"Testing f =", f;
"Group",  IdentifyGroup(GeometricAutomorphismGroupFromGenus3HyperellipticPolynomials(f));

// Generically, the 13 cov. of order 4 are non zeroes.
CoPrintAll(FdCov, f);

IO := Sort([[idx, FdCov[idx]`order] : idx in [1..#FdCov] |
    FdCov[idx]`order lt 8 and
    Degree(GetCovariant(FdCov[idx], FdCov, f)[1]) ge Max(1, FdCov[idx]`order-1) and
    Discriminant(GetCovariant(FdCov[idx], FdCov, f)[1]) ne 0 ],
    func<x, y | x[2]-y[2]>); 

IDX := [io : io in IO | io[2] eq 4];

// The cov. [ 12, 4 ], 
C := GetCovariant(FdCov[12], FdCov, f);
C; Factorization(Numerator(C[1]));

// It has a singularity only when a = +/2 or a = +/- 14 (a U6 curve) or a = 0 (V8 curve)
Factorization(Numerator(Discriminant(C[1])));

_<x> := PolynomialRing(Rationals());
f :=  (x^4-1)*(x^4 + 14*x^2 + 1);
IdentifyGroup(GeometricAutomorphismGroupFromGenus3HyperellipticPolynomials(f));

f :=  (x^4-1)*(x^4 - 14*x^2 + 1);
IdentifyGroup(GeometricAutomorphismGroupFromGenus3HyperellipticPolynomials(f));


// Few tests
_<X> := PolynomialRing(Rationals());
for i := 1 to 100 do
    repeat
	A := Random(-1000, 1000);
	g := (X^4-1)*(X^4 + A*X^2 +1);
    until IdentifyGroup(GeometricAutomorphismGroupFromGenus3HyperellipticPolynomials(g)) eq <8,2>;
    C := GetCovariant(FdCov[12], FdCov, g);
    ISOs := IsomorphismsWithCovariants(g, 8, C[1], C[3]);
//    #ISOs, "isomorphisms found";
    if #ISOs ne 4 then "ARGHHHH at", f; quit; end if;
end for;


/*********************************************************************************************/
/************************* Dimension 2 *******************************************************/
/*********************************************************************************************/


/* C2^3 (8 elts)
 ****************/
""; "**************** C2^3 test";

// Enumerating them
AF<a, b>  :=  AffineSpace(Rationals(), 2);
_<x> := PolynomialRing(Universe(Equations(Scheme(AF, a))));

f := (x^4 + a*x^2 + 1)*(x^4 + b*x^2 +1);
"Testing f =", f;
"Group",  IdentifyGroup(GeometricAutomorphismGroupFromGenus3HyperellipticPolynomials(f));

// Generically, the 13 cov. of order 4 are non zeroes.
CoPrintAll(FdCov, f);

IO := Sort([[idx, FdCov[idx]`order] : idx in [1..#FdCov] |
    FdCov[idx]`order lt 8 and
    Degree(GetCovariant(FdCov[idx], FdCov, f)[1]) ge Max(1, FdCov[idx]`order-1) and
    Discriminant(GetCovariant(FdCov[idx], FdCov, f)[1]) ne 0 ],
    func<x, y | x[2]-y[2]>); 

IDX := Sort([io : io in IO | io[2] eq 4]);

// The covs. 
EQS := []; for idx := 1 to 3 do
    C := GetCovariant(FdCov[IDX[idx,1]], FdCov, f);
    EQS cat:= [(Discriminant(C[1]))];
end for;

// It has a singularity only when
//               a + b = 0,                (C2xD8)
//               a*b - 2*a - 2*b - 12 = 0, (C2xD8)
//               a*b + 2*a + 2*b - 12 = 0, (C2xD8)

S := ReducedSubscheme(Scheme(AF, EQS));
IC := IrreducibleComponents(S);
for ic in IC do

    ic;
    
    QA<A,B> := quo<Universe(Equations(ic))  | Equations(ic)>;
    _<X> := PolynomialRing(QA);
    
    F := (X^4 + A*X^2 + 1)*(X^4 + B*X^2 +1);
    
    "Group",  IdentifyGroup(GeometricAutomorphismGroupFromGenus3HyperellipticPolynomials(F));
    "Singular", Discriminant(F) eq 0;
end for;


""; "Do these covariants have the good automorphism group ?";
EQS := C2p3Stratum;
for IJ in C2p3IJs do
    I, J := Explode(IJ);
    Append(~EQS, (4*I^3-J^2)*I*J);
end for;

SetVerbose("Groebner", 1);
GB := GroebnerBasis([ &*[f[1] : f in Factorization(EQS[i])]  : i in [1..#EQS] ]);

S := Scheme(AffineSpace(Parent(_J2)), GB);
RS := ReducedSubscheme(S);

SYS := Equations(RS);
[ Factorization(E) : E in SYS ];
#SYS;

ic := Scheme(AffineSpace(Parent(_J2)), SYS);

""; ic;

IDJ := ideal<Parent(_J2) | Equations(ic)>;
QJ := quo<Parent(_J2) | IDJ>;
FQJ := RingOfFractions(QJ);

j10, j9, j8, j7, j6, j5, j4, j3, j2 := Explode([FQJ.i : i in
    [1..Rank(FQJ)]]);
"Disc.",  DiscriminantFromShiodaInvariants([j2, j3, j4, j5, j6, j7, j8, j9, j10]);
"Group", GeometricAutomorphismGroupFromShiodaInvariants([j2, j3, j4, j5,
    j6, j7, j8, j9, j10]);


/* C4 (4 elts)
 ************/
""; "**************** C4 test";

// Enumerating them
AF<b, a>  :=  AffineSpace(Rationals(), 2);
_<x> := PolynomialRing(Universe(Equations(Scheme(AF, a))));


f := x*(x^2-1)*(x^4 + a*x^2 + b);
"Testing f =", f;
"Group",  IdentifyGroup(GeometricAutomorphismGroupFromGenus3HyperellipticPolynomials(f));

// Generically, the 13 cov. of order 4 are non zeroes.
CoPrintAll(FdCov, f);

IO := Sort([[idx, FdCov[idx]`order] : idx in [1..#FdCov] |
    FdCov[idx]`order lt 8 and
    Degree(GetCovariant(FdCov[idx], FdCov, f)[1]) ge Max(1, FdCov[idx]`order-1) and
    Discriminant(GetCovariant(FdCov[idx], FdCov, f)[1]) ne 0 ],
    func<x, y | x[2]-y[2]>); 
IDX := Sort([io : io in IO | io[2] eq 4]);

// The covs. 
EQS := []; for idx := 1 to 4 do
    C := GetCovariant(FdCov[IDX[idx,1]], FdCov, f);
    EQS cat:= [(Discriminant(C[1]))];
end for;


S := ReducedSubscheme(Scheme(AF, EQS));

IC := IrreducibleComponents(S);

for ic in IC do

    ic;
    
    QA<B, A> := quo<Universe(Equations(ic))  | Equations(ic)>;
    _<X> := PolynomialRing(QA);
    
    F := X * (X^2 - 1) * (X^4 + A*X^2 + B);
    
    "Group",  IdentifyGroup(GeometricAutomorphismGroupFromGenus3HyperellipticPolynomials(F));
    
end for;


/*********************************************************************************************/
/************************* Dimension 3 *******************************************************/
/*********************************************************************************************/


/* C2xC2 (4 elts)
 ****************/
""; "**************** C2xC2 test";

// Enumerating them
AF<c, a, b>  :=  AffineSpace(Rationals(), 3);
_<x> := PolynomialRing(Universe(Equations(Scheme(AF, a))));

f := x^8 + a*x^6 + b*x^4 + c*x^2 + 1;

"Testing f =", f;
GRPREF := IdentifyGroup(GeometricAutomorphismGroupFromGenus3HyperellipticPolynomials(f));
"Group", GRPREF;

// Generically, the 13 cov. of order 4 are non zeroes.
CoPrintAll(FdCov, f);

IO := Sort([[idx, FdCov[idx]`order] : idx in [1..#FdCov] |
    FdCov[idx]`order eq 4 and
    Degree(GetCovariant(FdCov[idx], FdCov, f)[1]) ge Max(1, FdCov[idx]`order-1) and
    Discriminant(GetCovariant(FdCov[idx], FdCov, f)[1]) ne 0 ],
    func<x, y | x[2]-y[2]>); 

// Generically, the 13 cov. of order 4 are non zeroes.
IDX := Sort([io : io in IO | io[2] eq 4]);


SingScheme := []; StillSingular := []; idxcur := 1;
C := GetCovariant(FdCov[IDX[idxcur,1]], FdCov, f);
S := ReducedSubscheme(Scheme(AF, Discriminant(C[1])));
IC := IrreducibleComponents(S);

for ic in IC do
    ic;
    QA<C, A, B> := quo<Universe(Equations(ic))  | Equations(ic)>;
    _<X> := PolynomialRing(QA);
    F := X^8 + A*X^6 + B*X^4 + C*X^2 + 1;
    GRP := IdentifyGroup(GeometricAutomorphismGroupFromGenus3HyperellipticPolynomials(F));
    "Group", GRP;
    if GRP eq GRPREF then
	Append(~StillSingular, ic);
    else
	Append(~SingScheme, ic);
    end if;
end for;
idxcur +:= 1;

"";
"Testing covariant ", idxcur;
    
NewSingular := [];
for S in StillSingular do
    EQS := Equations(S);
    
    "";
    "++++++ Based equations are", EQS;
    C1 := GetCovariant(FdCov[IDX[idxcur,1]], FdCov, f);
    C2 := GetCovariant(FdCov[IDX[idxcur+1,1]], FdCov, f);
    C3 := GetCovariant(FdCov[IDX[idxcur+2,1]], FdCov, f);
    RS := ReducedSubscheme(Scheme(AF, EQS cat [ Discriminant(C1[1]),
	                                        Discriminant(C2[1]),
	                                        Discriminant(C3[1]) ]));
    IC := IrreducibleComponents(RS);
	
    for ic in IC do
	"*** Testing", ic;
	QA<C, A, B> := quo<Universe(Equations(ic))  | Equations(ic)>;
	_<X> := PolynomialRing(QA);
	F := X^8 + A*X^6 + B*X^4 + C*X^2 + 1;
	GRP := IdentifyGroup(GeometricAutomorphismGroupFromGenus3HyperellipticPolynomials(F));
	"Group", GRP;
	if GRP eq GRPREF then
	    Append(~NewSingular, ic);
	else
	    Append(~SingScheme, ic);
	end if;
    end for;
    EQS;
end for;
StillSingular := NewSingular;
idxcur +:= 2;

while #StillSingular ne 0 do
    idxcur +:= 1;
    "";
    "Testing covariant ", idxcur;
    
    NewSingular := [];
    for S in StillSingular do
	EQS := Equations(S);

	"";
	"++++++ Based equations are", EQS;
	C := GetCovariant(FdCov[IDX[idxcur,1]], FdCov, f);
	RS := ReducedSubscheme(Scheme(AF, EQS cat [Discriminant(C[1])]));
	IC := IrreducibleComponents(RS);
	
	for ic in IC do
	    "*** Testing", ic;
	    QA<C, A, B> := quo<Universe(Equations(ic))  | Equations(ic)>;
	    _<X> := PolynomialRing(QA);
	    F := X^8 + A*X^6 + B*X^4 + C*X^2 + 1;
	    GRP := IdentifyGroup(GeometricAutomorphismGroupFromGenus3HyperellipticPolynomials(F));
	    "Group", GRP;
	    if GRP eq GRPREF then
		Append(~NewSingular, ic);
	    else
		Append(~SingScheme, ic);
	    end if;
	end for;
	EQS;
    end for;
    StillSingular := NewSingular;
end while;

// Tidying...
SG := [];
for idx := 1 to #SingScheme do
    Sc := SingScheme[idx];

    for sc in SG do
	if Sc subset sc then continue idx; end if;
    end for;

    for jdx := idx+1 to #SingScheme do
	sc := SingScheme[jdx];
	if Sc subset sc then continue idx; end if;
    end for;

    Append(~SG, Sc);
end for;

// Printing groups
for ic in SG do
    ic;
    QA<C, A, B> := quo<Universe(Equations(ic))  | Equations(ic)>;
    _<X> := PolynomialRing(QA);
    F := X^8 + A*X^6 + B*X^4 + C*X^2 + 1;
    "Group",  IdentifyGroup(GeometricAutomorphismGroupFromGenus3HyperellipticPolynomials(F));
end for;

""; "Do these covariants have the good automorphism group ?";
EQS := D4Stratum;
for IJ in D4IJs do
    I, J := Explode(IJ);
    Append(~EQS, (4*I^3-J^2)*I*J);
end for;

SetVerbose("Groebner", 1);
GB := GroebnerBasis([ &*[f[1] : f in Factorization(EQS[i])]  : i in [1..#EQS] ]);

S := Scheme(AffineSpace(Parent(_J2)), GB);
RS := ReducedSubscheme(S);

SYS := Equations(RS);
[ Factorization(E) : E in SYS ];
#SYS;

IC := IrreducibleComponents(Scheme(AffineSpace(Parent(_J2)), SYS));

SetVerbose("Groebner", 0);
for ic in IC do
    ""; ic;

    IDJ := ideal<Parent(_J2) | Equations(ic)>;
    QJ := quo<Parent(_J2) | IDJ>;
    FQJ := RingOfFractions(QJ);

    j10, j9, j8, j7, j6, j5, j4, j3, j2 := Explode([FQJ.i : i in
	[1..Rank(FQJ)]]);
    "Disc.",  DiscriminantFromShiodaInvariants([j2, j3, j4, j5, j6, j7, j8, j9, j10]);
    "Group", GeometricAutomorphismGroupFromShiodaInvariants([j2, j3, j4, j5,
	j6, j7, j8, j9, j10]);
end for;


/*********************************************************************************************/
/************************* Dimension 5 *******************************************************/
/*********************************************************************************************/

""; "**************** C2 test";

Relations := GroebnerBasis([Evaluate(EQ, [_J8, _J9, _J10]) :
    EQ in ShiodaAlgebraicInvariants([_J2, _J3, _J4, _J5, _J6, _J7] : ratsolve := false)]);

"Computing equations";
EQS := Relations;
for IJ in IJs do
    I, J := Explode(IJ);
    Append(~EQS, 4*I^3-J^2);
end for;


"Computing a Groebner basis";
SetVerbose("Groebner", 1);
GB := GroebnerBasis(EQS);
#GB;

GB := GroebnerBasis([ &*[f[1] : f in Factorization(GB[i])] : i in [1..#GB]]);

S := Scheme(AffineSpace(Parent(_J2)), GB);
RS := ReducedSubscheme(S);

SYS := Equations(RS);
[ Factorization(E) : E in SYS ];
#SYS;

IC := IrreducibleComponents(Scheme(AffineSpace(Parent(_J2)), SYS));

SetVerbose("Groebner", 0);
for ic in IC do
    ""; ic;

    IDJ := ideal<Parent(_J2) | Equations(ic)>;
    QJ := quo<Parent(_J2) | IDJ>;
    FQJ := RingOfFractions(QJ);

    j10, j9, j8, j7, j6, j5, j4, j3, j2 := Explode([FQJ.i : i in
	[1..Rank(FQJ)]]);
    "Disc.",  DiscriminantFromShiodaInvariants([j2, j3, j4, j5, j6, j7, j8, j9, j10]);
    "Group", GeometricAutomorphismGroupFromShiodaInvariants([j2, j3, j4, j5,
	j6, j7, j8, j9, j10]);
end for;

